from enum import Enum, auto


class CheckEnvFunc(Enum):
    """
    Functions used by FS_36_17_84_44 `check_env`.
    """

    func_id_get_server_project_git_commit_id = auto()
    """
    TODO
    """

    func_id_get_server_start_time = auto()
    """
    TODO
    """

    # TODO: add func to get server version
